from ..service import TaskService, AssigneeService, get_task_service, get_assignee_service
from .alconna import alc

from nonebot import require, logger
from nonebot.params import Depends
from nonebot.adapters import Event
from typing import Annotated
from uuid import UUID 

require("nonebot_plugin_saa")
from nonebot_plugin_saa import SaaTarget, MessageFactory

require("nonebot_plugin_orm")
from sqlalchemy.exc import NoResultFound, MultipleResultsFound

require("nonebot_plugin_alconna")
from nonebot_plugin_alconna import on_alconna, Arparma

rmd_app = on_alconna(alc)

@rmd_app.handle()
async def _(result: Arparma):
    if not result.matched:
        await rmd_app.send(f"命令解析失败。用法：\n{alc.get_help()}")
        await rmd_app.finish()


@rmd_app.assign("now")
async def rmd_now(saa_target: SaaTarget, task_service: Annotated[TaskService, Depends(get_task_service)]):
    task_ids = (await task_service.search_task(scope=saa_target)).scalars().all()
    for task_id in task_ids:
        await task_service.send_reminder(task_id)


@rmd_app.assign("add")
async def rmd_add(result: Arparma, saa_target: SaaTarget, task_service: Annotated[TaskService, Depends(get_task_service)]):
    logger.debug("Adding new task...")

    task_id = await task_service.create_task(
        name=result["add.task_name"],
        due_time=result["add.due_time"],
        remind_offset=result["remind_offset"],
        remind_interval=result["remind_interval"],
        recur_interval=result["recur_interval"],
        recur_type=result["recur_type"],
        platform_target=saa_target
    )
    logger.info(f"Task created: {task_id}")
    await (await task_service.describe_task(task_id)).send()
    logger.debug(f"Feedback sent")
    await rmd_app.finish()


@rmd_app.assign("rm")
async def rmd_delete(result: Arparma, saa_target: SaaTarget, task_service: Annotated[TaskService, Depends(get_task_service)]):
    try:
        task_id: UUID = (await task_service.search_task(result["rm.task_name"], saa_target)).scalar_one()
    except NoResultFound or MultipleResultsFound as e:
        logger.error(f"Unable to query task with name {result["rm.task_name"]}: {e}")
        await rmd_app.finish(f"查找任务\"{result["rm.task_name"]}\"时发生错误")
    else:
        await task_service.delete_task(task_id)
        await rmd_app.finish(f"任务\"{result["rm.task_name"]}\"已删除")


@rmd_app.assign("ls")
async def rmd_ls(saa_target: SaaTarget, task_service: Annotated[TaskService, Depends(get_task_service)]):
    task_ids = (await task_service.search_task(scope=saa_target)).scalars().all()
    msg = MessageFactory("当前会话中任务如下: ")
    logger.debug(f"Listing tasks... Task_ids: {task_ids}")
    for i in range(0, len(task_ids)):
        description = await task_service.describe_task(task_ids[i])
        msg += [f'\n{i+1}. '] + description
    await msg.send()
    await rmd_app.finish()


@rmd_app.assign("finish")
async def rmd_finish(result: Arparma, saa_target: SaaTarget, event: Event, task_service: Annotated[TaskService, Depends(get_task_service)]):
    logger.info(f"Finishing ... Lookup task {result["?task_name"]}")

    if not result["?task_name"]:
        logger.info(f"Try finish task in current chat for current user if one and only")
        logger.debug(f"current chat: {saa_target}, current user: {event.get_user_id()}")
        task_ids = (await task_service.search_task(scope=saa_target, user_id=event.get_user_id(), include_deleted=False)).scalars().all()
        if not task_ids:
            await rmd_app.finish("你在当前聊天下无任务")
        elif len(task_ids) > 1:
            msg = MessageFactory(["你在当前聊天下有多个任务, 请使用 '... finish <task_name> 指明要完成/跳过的任务'\n"])
            msg_count = 0
            for task_id in task_ids:
                msg_count += 1
                msg += [f"{msg_count}. ", await task_service.describe_task(task_id), " \n"]
            await rmd_app.finish()
        else:
            await task_service.finish_task(task_ids[0])
            await (await task_service.describe_task(task_ids[0])).send()
    else:
        try:
            task_id = (await task_service.search_task(task_name=result["?task_name"], scope=saa_target)).scalar_one()
        except NoResultFound:
            await rmd_app.finish(f"当前聊天下无“{result["?task_name"]}”任务")
        else:
            await task_service.finish_task(task_id)
            await ("任务已完成，下一次：" + await task_service.describe_task(task_id)).send()


@rmd_app.assign("skip")
async def rmd_skip(result: Arparma, saa_target: SaaTarget, event: Event, task_service: Annotated[TaskService, Depends(get_task_service)]):
    offset = result["offset"]
    logger.info(f"Skipping by {offset} ... Lookup task {result["?task_name"]}")

    if not result["?task_name"]:
        logger.info(f"Try skip task in current chat for current user if one and only")
        logger.debug(f"current chat: {saa_target}, current user: {event.get_user_id()}")
        task_ids = (await task_service.search_task(scope=saa_target, user_id=event.get_user_id(), include_deleted=False)).scalars().all()

        if not task_ids:
            await rmd_app.finish("你在当前聊天下无任务")
        elif len(task_ids) > 1:
            msg = MessageFactory(["你在当前聊天下有多个任务, 请使用 '... skip <task_name> 指明要完成/跳过的任务'\n"])
            msg_count = 0
            for task_id in task_ids:
                msg_count += 1
                msg += [f"{msg_count}. ", await task_service.describe_task(task_id), " \n"]
            await rmd_app.finish()
        else:
            await task_service.skip_task(task_ids[0], offset)
            await (await task_service.describe_task(task_ids[0])).send()
    else:
        try:
            task_id = (await task_service.search_task(task_name=result["?task_name"], scope=saa_target)).scalar_one()
        except NoResultFound:
            await rmd_app.finish(f"当前聊天下无“{result["?task_name"]}”任务")
        else:
            await task_service.skip_task(task_id, offset)
            await ("任务已跳过，下一次：" + await task_service.describe_task(task_id)).send()


@rmd_app.assign("due")
async def rmd_due(result: Arparma, saa_target: SaaTarget, task_service: Annotated[TaskService, Depends(get_task_service)]):
    task_id = (await task_service.search_task(task_name=result["due.task_name"], scope=saa_target)).scalar_one()
    if result["due_shift"]:
        await task_service.set_task(task_id, due_time=result["due_shift"])
    elif result["due_set"]:
        await task_service.set_task(task_id, due_time=result["due_set"])

    msg = await task_service.describe_due_time(task_id)
    await msg.send()


@rmd_app.assign("remind")
async def rmd_remind(result: Arparma, saa_target: SaaTarget, task_service: Annotated[TaskService, Depends(get_task_service)]):
    task_id = (await task_service.search_task(task_name=result["remind.task_name"], scope=saa_target)).scalar_one()

    if result["remind_offset"]:
        await task_service.set_task(task_id, remind_offset=result["remind_offset"])
    if result["remind_interval"]:
        await task_service.set_task(task_id, remind_interval=result["remind_interval"])

    msg = await task_service.describe_remind(task_id)
    await msg.send()


@rmd_app.assign("recur")
async def rmd_recur(result: Arparma, saa_target: SaaTarget, task_service: Annotated[TaskService, Depends(get_task_service)]):
    task_id = (await task_service.search_task(task_name=result["assign.task_name"], scope=saa_target)).scalar_one()

    if result["recur_type"]:
        await task_service.set_task(task_id, recur_type=result["recur_type"])
    if result["recur_interval"]:
        await task_service.set_task(task_id, recur_interval=result["recur_interval"])

    msg = await task_service.describe_recurrence(task_id)
    await msg.send()


@rmd_app.assign("assign")
async def rmd_assign(result: Arparma, saa_target: SaaTarget, task_service: Annotated[TaskService, Depends(get_task_service)], assignee_service: AssigneeService = Depends(get_assignee_service)):
    task_id = (await task_service.search_task(task_name=result["assign.task_name"], scope=saa_target)).scalar_one()
    assignee_ats = result["assign.?assignees"]

    if assignee_ats is not None:
        logger.info(f"Try adding / finding corresponding assignees to {assignee_ats}")
        user_ids = [at.target for at in assignee_ats]
        logger.debug(f"User ids: {user_ids}")
        assignee_ids = [(await assignee_service.add_assignee(user_id)) for user_id in user_ids]
        logger.debug(f"Assignee ids: {assignee_ids}")
        if result['options.rm']:
            await task_service.remove_assignments(task_id, assignee_ids)
        else:
            await task_service.create_assignments(task_id, assignee_ids)

    msg = await task_service.describe_assignee(task_id)
    await msg.send()



from enum import Enum as BuiltinEnum
from datetime import datetime, timedelta
import re
import pendulum


class RecurType(BuiltinEnum):
    Never = 0
    OnFinish = 1
    Regular = 2


def natural_lang_date(target_date: datetime, today=None):
    weekday_name = ["周一", "周二", "周三", "周四", "周五", "周六", "周日"]
    answer = ""

    if today is None:
        today = datetime.today()

    target_date = pendulum.instance(target_date)
    today = pendulum.instance(today)

    # 计算日期差（以天为单位）
    days_diff = (target_date.date() - today.date()).days

    # 如果日期相同
    if days_diff == 0:
        answer += "今天"
    elif days_diff == 1:
        answer += "明天"
    elif days_diff == -1:
        answer += "昨天"
    elif days_diff == 2:
        answer += "后天"
    elif days_diff == -2:
        answer += "前天"

    if answer != "":
        return answer + target_date.strftime("%H:%M")

    # 计算日期差的周数
    weeks_diff = (target_date - today).in_weeks()
    weekday = weekday_name[target_date.weekday()]

    # 处理在三周以内的日期
    if weeks_diff == 0:
        answer += f"本{weekday}"
    elif weeks_diff == 1:
        answer += f"下{weekday}"
    elif weeks_diff == 2:
        answer += f"两周后的{weekday}"
    elif weeks_diff == -1:
        answer += f"上{weekday}"
    elif weeks_diff == -2:
        answer += f"两周前的{weekday}"

    if answer != "":
        return answer + target_date.strftime("%H:%M")

    # 如果日期不在三周内，则返回标准日期格式
    return target_date.to_date_string() + target_date.strftime(" %H:%M")


def to_datetime(s: str) -> datetime:
    return pendulum.parse(s)


# From https://gist.github.com/santiagobasulto/698f0ff660968200f873a2f9d1c4113c

TIMEDELTA_REGEX = (r'((?P<days>-?\d+)d)?'
                   r'((?P<hours>-?\d+)h)?'
                   r'((?P<minutes>-?\d+)m)?'
                   r'((?P<seconds>-?\d+)s)?')
TIMEDELTA_PATTERN = re.compile(TIMEDELTA_REGEX, re.IGNORECASE)


def to_timedelta(s: str) -> timedelta:
    """ Parses a human-readable timedelta (3d5h19m1s) into a datetime.timedelta.
    Delta includes:
    * Xd days
    * Xh hours
    * Xm minutes
    * Xs seconds
    Values can be negative following timedelta's rules. Eg: -5h-30m
    """
    match = TIMEDELTA_PATTERN.match(s)
    if match:
        parts = {k: int(v) for k, v in match.groupdict().items() if v}
        return timedelta(**parts)


def to_recurtype(s: str) -> RecurType:
    return RecurType[s]



import uuid
from datetime import datetime, timedelta
from typing import Union, Set, Iterable

from apscheduler.triggers.interval import IntervalTrigger
from nonebot import require, logger
from collections.abc import AsyncGenerator

from .models import TaskModel, AssigneeModel, AssignmentModel, RecordModel
from .utils import natural_lang_date, RecurType

require("nonebot_plugin_apscheduler")
from nonebot_plugin_apscheduler import scheduler
from apscheduler.jobstores.base import JobLookupError

require("nonebot_plugin_saa")
from nonebot_plugin_saa import SaaTarget, MessageFactory, Mention, Text

require("nonebot_plugin_orm")
from nonebot_plugin_orm import get_session, async_scoped_session, AsyncSession, get_scoped_session
from sqlalchemy.future import select
from sqlalchemy.exc import NoResultFound, IntegrityError


def ensure_provided(ensure_args: list, not_none: int):
    if len([1 for arg in ensure_args if arg is not None]) != not_none:
        raise ValueError(f"Exactly {not_none} in specified range of optional argument(s) must be provided")


class Service:
    session: AsyncSession

    def __init__(self, session: AsyncSession):
        self.session = session

    async def __aenter__(self):
        # Optionally start a transaction if necessary
        # await self.session.begin()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if exc_type:
            await self.session.rollback()  # Rollback if an error occurred
        else:
            await self.session.commit()    # Commit if no exception
        await self.session.close()         # Close the session


class TaskService(Service):
    # Task related CRUD

    # TODO: a more general search function
    async def search_task(
            self,
            task_name: str | None = None,
            scope: SaaTarget | None = None,
            user_id: str | None = None,
            include_deleted: bool = False
    ):
        """Search for tasks with `task_name` in a scope of SaaTarget"""
        logger.debug(f"Searching tasks with the following filters: task_name={task_name}, scope={scope.model_dump_json()}, user_id={user_id}, include_deleted={include_deleted}")
        stmt = select(TaskModel.id)
        if task_name is not None:
            stmt = stmt.where(TaskModel.name == task_name)
        if scope is not None:
            stmt = stmt.where(TaskModel.platform_target_serial == scope.model_dump_json())
        if not include_deleted:
            stmt = stmt.where(TaskModel.is_deleted == False)
        if user_id is not None:
            stmt = (
                stmt
                .join(AssignmentModel, AssignmentModel.task_id == TaskModel.id)
                .join(AssigneeModel, AssigneeModel.id == AssignmentModel.assignee_id)
                .where(AssigneeModel.user_id == user_id)
            )
        logger.debug(f"Querying with statement: \n{stmt}")
        results = await self.session.execute(stmt)
        return results

    async def __get_task(
            self,
            task_id: Union[uuid.UUID, None] = None,
            include_deleted: bool = False
    ):
        """Fetch a task by ID, """
        stmt = select(TaskModel)
        if task_id:
            stmt = stmt.where(TaskModel.id == task_id)
        if not include_deleted:
            stmt = stmt.where(TaskModel.is_deleted == False)

        try:
            result = await self.session.execute(stmt)
            return result.scalar_one()
        except NoResultFound:
            logger.error(f"Task with ID {task_id} not found.")
            raise NoResultFound(f"No {"undeleted " if not include_deleted else ""}task has id {task_id}")

    async def create_task(
            self,
            name: str,
            due_time: datetime,
            remind_offset: timedelta,
            remind_interval: timedelta,
            recur_interval: timedelta,
            recur_type: RecurType,
            platform_target: SaaTarget
    ) -> uuid.UUID:
        """Create task"""
        created_task = TaskModel(
            name=name,
            due_time=due_time,
            remind_offset=remind_offset,
            remind_interval=remind_interval,
            recur_interval=recur_interval,
            recur_type=recur_type,
            platform_target_serial=platform_target.model_dump_json(),
            current_assignment_order=None
        )
        self.session.add(created_task)
        await self.session.commit()
        await self.session.refresh(created_task)
        await self.refresh_reminder(created_task.id)
        task_id = created_task.id

        return task_id

    async def delete_task(self, task_id: uuid.UUID) -> None:
        """Delete given task or task with given id"""
        await self.remove_reminder(task_id)
        task = await self.__get_task(task_id=task_id)
        task.soft_delete()
        await self.session.commit()

    async def set_task(
            self,
            task_id: Union[uuid.UUID, None] = None,
            **kwargs
    ) -> None:
        """Set attributes of a task"""
        logger.debug(f"Setting values: {kwargs}")
        task = await self.__get_task(task_id)
        for attr_name, attr_value in kwargs.items():
            if not hasattr(task, attr_name):
                logger.error(f"Setting non-existing attr {attr_name} of task {task.id} to {attr_value}")
                raise ValueError(f"None existing attr: {attr_name}")

            match attr_name:
                case "due_time":
                    if isinstance(attr_value, timedelta):
                        setattr(task, "due_time", getattr(task, "due_time") + attr_value)
                    elif isinstance(attr_value, datetime):
                        setattr(task, attr_name, attr_value)
                    else:
                        raise ValueError("Incorrect value type for due_time")
                    await self.refresh_reminder(task.id)
                case "remind_offset" | "remind_interval":
                    setattr(task, attr_name, attr_value)
                    await self.refresh_reminder(task.id)
                case _:
                    setattr(task, attr_name, attr_value)

        await self.session.commit()
        await self.session.refresh(task)

    # Assignee lookup & assignment CRUD
    async def get_assignee_user_ids(self, task_id: uuid.UUID):
        """Get all assignees for a task."""
        assignees = (
            await self.session.execute(
                select(AssigneeModel.user_id)
                .join(AssignmentModel, AssignmentModel.assignee_id == AssigneeModel.id)
                .join(TaskModel, TaskModel.id == AssignmentModel.task_id)
                .where(TaskModel.id == task_id)
                .order_by(AssignmentModel.order)
            )
        ).scalars().all()
        if not assignees:
            logger.info(f"No assignee for task {task_id}")
            return []
        logger.debug(f"Assignees for task {task_id}: {assignees}")
        return assignees

    async def shift_current_assignee_order(self, task_id: uuid.UUID, offset: int):
        """Shift by offset to another from current assignee"""
        logger.info(f"Shifting task current by {offset}")
        assignee_count = len(await self.get_assignee_user_ids(task_id))
        if assignee_count == 0:
            logger.warning("Shifting task current with no assignee")
        task = await self.__get_task(task_id)
        logger.debug(f"Previous order: {task.current_assignment_order}")
        task.current_assignment_order = (task.current_assignment_order + offset) % assignee_count
        logger.debug(f"Assignee count: {assignee_count}, current order: {task.current_assignment_order}")

        await self.session.commit()

    async def create_assignments(self, task_id: uuid.UUID, assignee_ids: Iterable[uuid.UUID]) -> None:
        """Create assignments."""
        logger.info(f"Assigning task {task_id} to assinees {assignee_ids}")
        task = await self.__get_task(task_id)

        assignee_count = len(await self.get_assignee_user_ids(task_id))
        if not task.current_assignment_order:
            logger.debug(f"Initialize task's current assignment order to 0")
            task.current_assignment_order = 0
        await self.session.commit()
        await self.session.refresh(task)
        logger.debug(f"Existing assignee count is {assignee_count} for task {task_id}")

        for assignee_id in assignee_ids:
            new_assignment = AssignmentModel(
                task_id=task_id,
                assignee_id=assignee_id,
                order=assignee_count
            )
            self.session.add(new_assignment)
            try:
                await self.session.commit()
            except IntegrityError as e:
                await self.session.rollback()
                logger.warning(f"Assignee {assignee_id} already assigned to task {task_id}")
            else:
                logger.debug(f"Assignees added for task {task_id}: {assignee_id} [{assignee_count}]")
                assignee_count += 1

    async def remove_assignments(self, task_id: uuid.UUID, assignee_ids: Iterable[uuid.UUID]):
        """Try remove assignment from given task"""
        task = await self.__get_task(task_id)

        assignments = (
            await self.session.execute(
                select(AssignmentModel)
                .where(AssignmentModel.id.in_(assignee_ids))
                .order_by(AssignmentModel.order)
            )
        ).scalars().all()

        logger.debug(f"Assignments for task {task.name}: {assignments}")

        offset = 0
        assignee_ids = list(assignee_ids)

        for i in range(0, len(assignments)):
            if assignments[i].assignee_id in assignee_ids:
                assignee_ids.remove(assignments[i].assignee_id)
                self.session.delete(assignments[i])
                offset += 1
                if task.current_assignment_order >= i: task.current_assignment_order -= 1
            else:
                assignments[i].order -= offset

        if assignee_ids:
            logger.warning(f"These assignees is not assigned to task {task.id}: {assignee_ids} ")

    # APScheduler wakeup timer related

    # Note: as apscheduler handles its own persist database, this operation is not atomic: 
    # a partial fail could lead to wild wakeup jobs waking up a task more frequently than 
    # expected, or trying to wake up a task no longer exist. The purge method in utils could
    # fix all such errors but is relatively expensive.

    @staticmethod
    async def send_reminder(task_id: uuid.UUID):
        """Send a reminder notification for the task."""
        async with TaskService(get_session()) as task_service:
            logger.info(f"Sending notification for task {task_id}")
            try:
                task = await task_service.__get_task(task_id)
            except NoResultFound:
                logger.error(f"No active task with id {task_id}. Possibly unmanaged jobs exist, please purge.")
                return
            msg = await task_service.get_notification_message(task_id)
            try:
                await msg.send_to(target=task.platform_target)
                logger.debug(f"Sent reminder for task {task.id}")
            except Exception as e:
                logger.error(f"Failed to send reminder for task {task.id}: {e}")
                return

    @staticmethod
    async def send_reminder_for_all(scope: Set[SaaTarget]):
        """Send a reminder notification for all ongoing task."""
        session = get_session()
        scope_serialized = {saa_target.model_dump_json() for saa_target in scope}
        task_ids = (
            await session.execute(
                select(TaskModel.id)
                .where(
                    TaskModel.is_deleted == False,
                    TaskModel.platform_target_serial.in_(scope_serialized)
                )
            )
        ).scalars().all()
        for task_id in task_ids:
            await TaskService.send_reminder(task_id)

    async def schedule_reminder(self, task_id: uuid.UUID):
        """Schedule the APScheduler reminder job for the task."""
        task = await self.__get_task(task_id)
        trigger = IntervalTrigger(
            seconds=int(task.remind_interval.total_seconds()),
            jitter=int(task.remind_interval.total_seconds() * 0.02),
            start_date=task.due_time + task.remind_offset
        )
        job = scheduler.add_job(
            TaskService.send_reminder,
            jobstore='nonebot-plugin-yareminder-jobstore',
            trigger=trigger,
            args=[task.id],
            name=f"Reminder wakeup timer for task {task.name} ({task.id})"
        )
        task.apscheduler_job_id = job.id
        await self.session.commit()
        await self.session.refresh(task)
        logger.debug(f"Scheduled reminder for task {task.id}: {job.id}")

    async def remove_reminder(self, task_id: uuid.UUID):
        """Remove the APScheduler reminder job."""
        task = await self.__get_task(task_id)
        if task.apscheduler_job_id:
            try:
                scheduler.remove_job(task.apscheduler_job_id, 'nonebot-plugin-yareminder-jobstore')
                logger.debug(f"Removed reminder for task {task.id}")
            except JobLookupError:
                logger.warning(f"Job {task.apscheduler_job_id} not found for task {task.id}")
            task.apscheduler_job_id = None
            await self.session.commit()
            await self.session.refresh(task)
        else:
            logger.warning(f"Task {task_id} has no job")

    async def refresh_reminder(self, task_id: uuid.UUID):
        """Refresh the reminder by removing and rescheduling the APScheduler job."""
        logger.debug(f"Refreshing reminder for task {task_id}")

        await self.remove_reminder(task_id)
        await self.schedule_reminder(task_id)

    # All human-readable related message generation

    async def get_notification_message(self, task_id: uuid.UUID) -> MessageFactory:
        """Generate the notification message for the task."""
        msg = MessageFactory()
        task: TaskModel = await self.__get_task(task_id)

        assignee_user_ids = await self.get_assignee_user_ids(task_id)
        if assignee_user_ids:
            msg += [Mention(user_id=assignee_user_ids[task.current_assignment_order]), " "]
        now = datetime.now()
        if now < task.due_time:
            msg += ["请记得", await self.describe_due_time(task_id), str(task.name)]
        else:
            msg += [f"{task.name}应", await self.describe_due_time(task_id), "哦"]

        return msg

    async def describe_recurrence(self, task_id: uuid.UUID) -> Text:
        """Return a Text that describes the recurrence """
        task = await self.__get_task(task_id)
        match task.recur_type:
            case RecurType.Never:
                return Text("不重复")
            case RecurType.OnFinish:
                return Text(f"完成{task.recur_interval}后重复")
            case RecurType.Regular:
                return Text(f"每{task.recur_interval}重复")

    async def describe_due_time(self, task_id: uuid.UUID) -> Text:
        """Return a Text that describes the recurrence """
        task = await self.__get_task(task_id)
        return Text(f"在{natural_lang_date(task.due_time)}前完成")

    async def describe_remind(self, task_id: uuid.UUID) -> Text:
        """Returns a Text that describes the reminder offset"""
        task = await self.__get_task(task_id)
        return Text(f"提前{task.remind_offset}提醒，提醒间隔{task.remind_interval}")

    async def describe_assignee(self, task_id: uuid.UUID) -> MessageFactory:
        """Returns a MessageFactory that describes the assignees and current one"""
        msg = MessageFactory()
        task = await self.__get_task(task_id)
        user_ids = await self.get_assignee_user_ids(task.id)
        if user_ids:
            logger.debug(f"Exist assignees for task {task_id}")
            msg += "依次由 "
            for user_id in user_ids:
                msg += Mention(user_id)
                msg += " "
            msg += " 完成，当前轮到 "
            msg += Mention(user_ids[task.current_assignment_order])
        else:
            logger.debug(f"No assignees for task {task_id}")
            msg += "无指派"
        return msg

    async def describe_task(
            self,
            task_id: uuid.UUID
    ):
        logger.debug(f"Generating description of task: {task_id}")
        task: TaskModel = await self.__get_task(task_id, include_deleted=True)
        msg = MessageFactory()
        if task.is_deleted:
            msg += f"[{task.name}] 已被删除"
        else:
            msg += f"[{task.name}] "
            msg += (await self.describe_due_time(task_id)) + "，"
            msg += (await self.describe_remind(task_id))+ "，"
            msg += (await self.describe_recurrence(task_id))+ "，"
            msg += (await self.describe_assignee(task_id))

        return msg

    # Task Normal behaviour related
    # include: 
    # - finish (change due time and current assignee, and reschedule remind job)
    # - skip (finish without due time change or rescheduler)

    async def finish_task(self, task_id: uuid.UUID):
        """Mark a task as finished and reschedule if it has recurring intervals."""
        task = await self.__get_task(task_id)
        try:
            assignment = (
                await self.session.execute(
                    select(AssignmentModel)
                    .where(
                        AssignmentModel.task_id == task.id,
                        AssignmentModel.order == task.current_assignment_order
                    )
                )
            ).scalar_one()
        except NoResultFound:
            logger.info(f"No assignment for task {task_id}")
            assignee_id = None
        else:
            assignee_id = assignment.assignee_id

        new_record = RecordModel(
            task_id=task_id,
            assignee_id=assignee_id,
            due_time=task.due_time,
            finish_time=datetime.now()
        )

        self.session.add(new_record)

        match task.recur_type:
            case RecurType.Never:
                await self.remove_reminder(task_id)
                task.due_time = None
                await self.delete_task(task)
            case RecurType.OnFinish:
                task.due_time = datetime.now() + task.recur_interval
                await self.refresh_reminder(task_id)
                if assignee_id is not None:
                    await self.shift_current_assignee_order(task_id, 1)
            case RecurType.Regular:
                task.due_time = task.due_time + task.recur_interval
                await self.refresh_reminder(task_id)
                if assignee_id is not None:
                    await self.shift_current_assignee_order(task_id, 1)

        await self.session.commit()
        await self.session.refresh(task)
        logger.debug(
            f"Finished task {task.id}, next due: {task.due_time}, next assignee: {task.current_assignment_order}")

    async def skip_task(self, task_id: uuid.UUID, offset: int):
        """Mark a task as finished and reschedule if it has recurring intervals."""
        task = await self.__get_task(task_id)

        if len(await self.get_assignee_user_ids(task_id)) <= 1:
            logger.warning(f"Task {task_id} with one or none assignee cannot be skipped")
            return

        await self.shift_current_assignee_order(task_id, offset)

        await self.session.commit()
        await self.session.refresh(task)
        logger.debug(
            f"Skipped task {task.id}, next due: {task.due_time}, next assignee: {task.current_assignment_order}")

    async def purge_wild_jobs(self) -> int:
        # obtain all job id from job store
        apscheduler_job_ids = [job.id for job in scheduler.get_jobs()]
        count = 0

        controlled_job_ids = (
            await self.session.execute(
                select(TaskModel.apscheduler_job_id)
                .where(TaskModel.is_deleted == False)
            )
        ).scalars().all()

        for job_id in apscheduler_job_ids:
            if job_id not in controlled_job_ids:
                scheduler.remove_job(job_id)
                count += 1

        return count


class AssigneeService(Service):
    async def add_assignee(self, user_id: str):
        """Add an assignee"""
        logger.debug(f"Adding assignee {user_id}")
        new_assignee = AssigneeModel(user_id=user_id)
        self.session.add(new_assignee)

        try:
            await self.session.commit()
        except IntegrityError as e:
            logger.debug(f"Duplication detected: {e}, try locating existing assignee id")
            await self.session.rollback()

            return (
                await self.session.execute(
                    select(AssigneeModel.id)
                    .where(AssigneeModel.user_id == user_id)
                )
            ).scalar_one()

        await self.session.refresh(new_assignee)
        return new_assignee.id


async def get_task_service() -> AsyncGenerator[TaskService, None]:
    session = get_scoped_session()
    async with TaskService(session) as task_service:
        yield task_service


async def get_assignee_service() -> AsyncGenerator[TaskService, None]:
    session = get_scoped_session()
    async with AssigneeService(session) as assignee_service:
        yield assignee_service



from ..utils import to_datetime, to_timedelta, to_recurtype, RecurType

from nonebot import require
from datetime import timedelta

require("nonebot_plugin_alconna")
from nonebot_plugin_alconna import At
from arclet.alconna import Arg, Alconna, Option, MultiVar, Subcommand

alc = Alconna(
    ":rmd",
    Subcommand("now"),
    Subcommand(
        "add",
        Arg("task_name", str),
        Arg("due_time", to_datetime),
        Option("-i|--remind-interval", Arg("remind_interval", to_timedelta, timedelta(hours=3))),
        Option("-o|--remind-offset", Arg("remind_offset", to_timedelta, timedelta(hours=-24))),
        Option("-t|--recur-type", Arg("recur_type", to_recurtype, RecurType.Regular)),
        Option("-r|--recur-interval", Arg("recur_interval", to_timedelta, timedelta(days=2)))
    ),
    Subcommand("rm", Arg("task_name", str)),
    Subcommand("ls"),
    Subcommand("finish", Arg("?task_name", str)),
    Subcommand("skip", Arg("?task_name", str), Option("--offset", Arg("offset", int, 1))),
    Subcommand(
        "due",
        Arg("task_name", str),
        Option("--shift", Arg("due_shift", to_timedelta)),
        Option("--set", Arg("due_set", to_datetime))
    ),
    Subcommand(
        "remind",
        Arg("task_name", str),
        Option("-o|--offset", Arg("remind_offset", to_timedelta)),
        Option("-i|--interval", Arg("remind_interval", to_timedelta))
    ),
    Subcommand(
        "recur",
        Arg("task_name", str),
        Option("-t|--type", Arg("recur_type", to_recurtype)),
        Option("-i|--interval", Arg("recur_interval", to_timedelta))
    ),
    Subcommand(
        "assign",
        Arg("task_name", str),
        Option("-r|--rm"),
        Arg("?assignees", MultiVar(At))
    )
)



from .utils import RecurType

from nonebot import require
from datetime import datetime, timedelta
import uuid

require("nonebot_plugin_orm")
from nonebot_plugin_orm import Model
from sqlalchemy import ForeignKey, Column, Integer, String, DateTime, Interval, JSON, Boolean, Enum, UniqueConstraint
from sqlalchemy.orm import Mapped, mapped_column
from sqlalchemy.ext.declarative import declared_attr

require("nonebot_plugin_saa")
from nonebot_plugin_saa import PlatformTarget


class SoftDeleteMixin:
    is_deleted: bool
    deleted_at: datetime | None

    @declared_attr
    def is_deleted(self):
        return Column(Boolean, default=False)

    @declared_attr
    def deleted_at(self):
        return Column(DateTime, default=datetime.min)

    def soft_delete(self):
        self.is_deleted = True
        self.deleted_at = datetime.now()

    def restore(self):
        self.is_deleted = False
        self.deleted_at = None


class TaskModel(Model, SoftDeleteMixin):
    # trival attributes:
    # name, description, recur_type, recur_interval: no hook on change
    # untrival attributes:
    # due_time / remind_offset / remind_interval: delete all remind timers and reschedule
    id: Mapped[uuid.UUID] = mapped_column(String(36), primary_key=True, default=lambda: str(uuid.uuid4()),
                                          nullable=False)
    name: Mapped[str] = mapped_column(String)
    due_time: Mapped[datetime] = mapped_column(DateTime)
    remind_offset: Mapped[timedelta] = mapped_column(Interval)
    remind_interval: Mapped[timedelta] = mapped_column(Interval)
    recur_interval: Mapped[timedelta] = mapped_column(Interval, nullable=True)
    recur_type: Mapped[RecurType] = mapped_column(Enum(RecurType))
    apscheduler_job_id: Mapped[str] = mapped_column(String, nullable=True)
    platform_target_serial: Mapped[str] = mapped_column(String, nullable=True)
    current_assignment_order: Mapped[int] = mapped_column(Integer, nullable=True, default=None)

    __table_args__ = (
        UniqueConstraint(
            'name',
            'platform_target_serial',
            'is_deleted',
            'deleted_at',
            name='unique_name_in_the_same_session_with_undeleted_tasks'
        ),
    )

    @property
    def platform_target(self):
        return PlatformTarget.deserialize(self.platform_target_serial)

    @platform_target.setter
    def platform_target(self, platform_target: PlatformTarget):
        self.platform_target_serial = platform_target.model_dump()

    def __repr__(self) -> str:
        return (
            f"ID: {self.id}, Name: {self.name}, "
            f"Due: {self.due_time}, Current: {self.current_assignment_order}, Remind offset: {self.remind_offset}, "
            f"Remind interval: {self.remind_interval}, Recur interval: {self.recur_interval},"
            f"Recur Type: {self.recur_type},"
            f"Apscheduler ID: {self.apscheduler_job_id}, Deleted time: {self.deleted_at}"
        )


class AssigneeModel(Model):
    id: Mapped[uuid.UUID] = mapped_column(String(36), primary_key=True, default=lambda: str(uuid.uuid4()),
                                          nullable=False)
    user_id: Mapped[str] = mapped_column(String, unique=True)


class AssignmentModel(Model):
    id: Mapped[uuid.UUID] = mapped_column(String(36), primary_key=True, default=lambda: str(uuid.uuid4()),
                                          nullable=False)
    task_id: Mapped[uuid.UUID] = mapped_column(ForeignKey(TaskModel.__tablename__ + ".id"))
    assignee_id: Mapped[uuid.UUID] = mapped_column(ForeignKey(AssigneeModel.__tablename__ + ".id"))
    order: Mapped[int] = mapped_column(Integer)

    __table_args__ = (
        UniqueConstraint('task_id', 'order', name='different_seq for task'),
        UniqueConstraint('task_id', 'assignee_id', name='can only assign once'),
    )


class RecordModel(Model):
    id: Mapped[uuid.UUID] = mapped_column(String(36), primary_key=True, default=lambda: str(uuid.uuid4()),
                                          nullable=False)
    task_id: Mapped[uuid.UUID] = mapped_column(ForeignKey(TaskModel.__tablename__ + ".id"))
    assignee_id: Mapped[uuid.UUID] = mapped_column(ForeignKey(TaskModel.__tablename__ + ".id"), nullable=True)
    due_time: Mapped[datetime] = mapped_column(DateTime)
    finish_time: Mapped[datetime] = mapped_column(DateTime)



from nonebot import require
from nonebot.plugin import PluginMetadata

from .command import rmd_app

__plugin_meta__ = PluginMetadata(
    name="nonebot-plugin-yareminder",
    description="Yet another reminder plugin supporting per-task configuration and check-in.",
    usage=":rmd COMMAND [ARGS] [OPTIONS]",
    type="application",
)

require('nonebot_plugin_saa')
from nonebot_plugin_saa import enable_auto_select_bot
enable_auto_select_bot()

require("nonebot_plugin_localstore")
from nonebot_plugin_localstore import get_data_dir

require("nonebot_plugin_apscheduler")
from nonebot_plugin_apscheduler import scheduler
from apscheduler.jobstores.sqlalchemy import SQLAlchemyJobStore
scheduler.add_jobstore(
    SQLAlchemyJobStore(url="sqlite:///" + str(get_data_dir(__plugin_meta__.name) / "apscheduler.sqlite3")),
    alias='nonebot-plugin-yareminder-jobstore'
)
